@@ -1,9 +1,14 @@
+{-# LANGUAGE DerivingVia #-}
 {-# LANGUAGE NoMonoLocalBinds #-}
 {-# LANGUAGE NoMonomorphismRestriction #-}
 
 module Bluefin.Internal.Examples where
 
 import Bluefin.Internal hiding (b, w)
+import Bluefin.Internal.OneWayCoercible
+  ( OneWayCoercible (oneWayCoercibleImpl),
+    gOneWayCoercible,
+  )
 import Bluefin.Internal.Pipes
   ( Producer,
     runEffect,
@@ -20,6 +25,7 @@ import Control.Monad.IO.Class (liftIO)
 import Data.Foldable (for_)
 import Data.Monoid (Any (Any, getAny))
 import Data.Proxy (Proxy (Proxy))
+import GHC.Generics (Generic)
 import Text.Read (readMaybe)
 import Prelude hiding
   ( break,
@@ -663,22 +669,20 @@ exampleCounter4 = runPureEff $ yieldToList $ \y -> do
 -- Counter 5
 
 data Counter5 e = MkCounter5
-  { incCounter5Impl :: forall e'. Eff (e' :& e) (),
-    getCounter5Impl :: forall e'. String -> Eff (e' :& e) Int
+  { incCounter5Impl :: Eff e (),
+    getCounter5Impl :: String -> Eff e Int
   }
+  deriving (Generic)
+  deriving (Handle) via OneWayCoercibleHandle Counter5
 
-instance Handle Counter5 where
-  mapHandle c =
-    MkCounter5
-      { incCounter5Impl = useImplUnder (incCounter5Impl c),
-        getCounter5Impl = \msg -> useImplUnder (getCounter5Impl c msg)
-      }
+instance (e :> es) => OneWayCoercible (Counter5 e) (Counter5 es) where
+  oneWayCoercibleImpl = gOneWayCoercible
 
 incCounter5 :: (e :> es) => Counter5 e -> Eff es ()
-incCounter5 e = makeOp (incCounter5Impl (mapHandle e))
+incCounter5 e = incCounter5Impl (mapHandle e)
 
 getCounter5 :: (e :> es) => Counter5 e -> String -> Eff es Int
-getCounter5 e msg = makeOp (getCounter5Impl (mapHandle e) msg)
+getCounter5 e msg = getCounter5Impl (mapHandle e) msg
 
 runCounter5 ::
   (e1 :> es) =>
@@ -723,21 +727,18 @@ exampleCounter5 = runPureEff $ yieldToList $ \y -> do
 -- Counter 6
 
 data Counter6 e = MkCounter6
-  { incCounter6Impl :: forall e'. Eff (e' :& e) (),
+  { incCounter6Impl :: Eff e (),
     counter6State :: State Int e,
     counter6Stream :: Stream String e
   }
+  deriving (Generic)
+  deriving (Handle) via OneWayCoercibleHandle Counter6
 
-instance Handle Counter6 where
-  mapHandle c =
-    MkCounter6
-      { incCounter6Impl = useImplUnder (incCounter6Impl c),
-        counter6State = mapHandle (counter6State c),
-        counter6Stream = mapHandle (counter6Stream c)
-      }
+instance (e :> es) => OneWayCoercible (Counter6 e) (Counter6 es) where
+  oneWayCoercibleImpl = gOneWayCoercible
 
 incCounter6 :: (e :> es) => Counter6 e -> Eff es ()
-incCounter6 e = makeOp (incCounter6Impl (mapHandle e))
+incCounter6 e = incCounter6Impl (mapHandle e)
 
 getCounter6 :: (e :> es) => Counter6 e -> String -> Eff es Int
 getCounter6 (MkCounter6 _ st y) msg = do
@@ -792,7 +793,7 @@ data Counter7 e = MkCounter7
   }
 
 instance Handle Counter7 where
-  mapHandle c =
+  handleImpl = handleMapHandle $ \c ->
     MkCounter7
       { incCounter7Impl = \ex -> useImplUnder (incCounter7Impl c ex),
         counter7State = mapHandle (counter7State c),
@@ -860,23 +861,21 @@ exampleCounter7B = runPureEff $ yieldToList $ \y -> do
 -- FileSystem
 
 data FileSystem es = MkFileSystem
-  { readFileImpl :: forall e. FilePath -> Eff (e :& es) String,
-    writeFileImpl :: forall e. FilePath -> String -> Eff (e :& es) ()
+  { readFileImpl :: FilePath -> Eff es String,
+    writeFileImpl :: FilePath -> String -> Eff es ()
   }
+  deriving (Generic)
+  deriving (Handle) via OneWayCoercibleHandle FileSystem
 
-instance Handle FileSystem where
-  mapHandle fs =
-    MkFileSystem
-      { readFileImpl = \fp -> useImplUnder (readFileImpl fs fp),
-        writeFileImpl = \fp s -> useImplUnder (writeFileImpl fs fp s)
-      }
+instance (e :> es) => OneWayCoercible (FileSystem e) (FileSystem es) where
+  oneWayCoercibleImpl = gOneWayCoercible
 
 readFile :: (e :> es) => FileSystem e -> FilePath -> Eff es String
-readFile fs filepath = makeOp (readFileImpl (mapHandle fs) filepath)
+readFile fs filepath = readFileImpl (mapHandle fs) filepath
 
 writeFile :: (e :> es) => FileSystem e -> FilePath -> String -> Eff es ()
 writeFile fs filepath contents =
-  makeOp (writeFileImpl (mapHandle fs) filepath contents)
+  writeFileImpl (mapHandle fs) filepath contents
 
 runFileSystemPure ::
   (e1 :> es) =>
@@ -948,23 +947,15 @@ exampleRunFileSystemIO = runEff_ $ \io -> try $ \ex ->
 -- instance Handle example
 
 data Application e = MkApplication
-  { queryDatabase :: forall e'. String -> Int -> Eff (e' :& e) [String],
+  { queryDatabase :: String -> Int -> Eff e [String],
     applicationState :: State (Int, Bool) e,
     logger :: Stream String e
   }
+  deriving (Generic)
+  deriving (Handle) via OneWayCoercibleHandle Application
 
-instance Handle Application where
-  mapHandle
-    MkApplication
-      { queryDatabase = q,
-        applicationState = a,
-        logger = l
-      } =
-      MkApplication
-        { queryDatabase = \s i -> useImplUnder (q s i),
-          applicationState = mapHandle a,
-          logger = mapHandle l
-        }
+instance (e :> es) => OneWayCoercible (Application e) (Application es) where
+  oneWayCoercibleImpl = gOneWayCoercible
 
 -- This example shows a case where we can use @bracket@ polymorphically
 -- in order to perform correct cleanup if @es@ is instantiated to a
@@ -1046,14 +1037,14 @@ rethrowIOExample = runEff_ $ \io -> do
     Right contents -> contents
 
 data DynamicReader r e = DynamicReader
-  { askLRImpl :: forall e'. Eff (e' :& e) r,
+  { askLRImpl :: Eff e r,
     localLRImpl :: forall e' a. (r -> r) -> Eff e' a -> Eff (e' :& e) a
   }
 
 instance Handle (DynamicReader r) where
-  mapHandle h =
+  handleImpl = handleMapHandle $ \h ->
     DynamicReader
-      { askLRImpl = useImplUnder (askLRImpl h),
+      { askLRImpl = useImpl (askLRImpl h),
         localLRImpl = \f k -> useImplUnder (localLRImpl h f k)
       }
 
@@ -1061,7 +1052,7 @@ askLR ::
   (e :> es) =>
   DynamicReader r e ->
   Eff es r
-askLR c = makeOp (askLRImpl (mapHandle c))
+askLR c = askLRImpl (mapHandle c)
 
 localLR ::
   (e :> es) =>
@@ -1081,7 +1072,7 @@ runDynamicReader r k =
       k
       DynamicReader
         { askLRImpl = ask h,
-          localLRImpl = \f k' -> makeOp (local h f (useImpl k'))
+          localLRImpl = \f k' -> local h f (useImpl k')
         }
 
 -- Fails to compile unless '(e :> es) => e :> (x :& es)' is incoherent
@@ -1091,14 +1082,14 @@ example = runPureEff $
   evalState () $ \st1 ->
     evalState () $ \st2 -> do
       Proxy :: Proxy es <- effTag
-      subset @es @es
+      satisfied @(es :> es)
 
       Proxy :: Proxy e1 <- handleTag st1
       Proxy :: Proxy e2 <- handleTag st2
 
-      subset @e1 @e1
-      subset @e2 @e2
-      subset @e1 @(e1 :& e2)
-      subset @e2 @(e1 :& e2)
+      satisfied @(e1 :> e1)
+      satisfied @(e2 :> e2)
+      satisfied @(e1 :> (e1 :& e2))
+      satisfied @(e2 :> (e1 :& e2))
 
-      subset @(e1 :& e2) @(e1 :& e2)
+      satisfied @((e1 :& e2) :> (e1 :& e2))
